/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.xwiki.refactoring.internal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Provider;
import javax.inject.Singleton;

import org.apache.commons.lang3.exception.ExceptionUtils;
import org.slf4j.Logger;
import org.xwiki.component.annotation.Component;
import org.xwiki.job.api.AbstractCheckRightsRequest;
import org.xwiki.job.event.status.JobProgressManager;
import org.xwiki.link.LinkStore;
import org.xwiki.model.EntityType;
import org.xwiki.model.reference.DocumentReference;
import org.xwiki.model.reference.DocumentReferenceResolver;
import org.xwiki.model.reference.EntityReference;
import org.xwiki.model.reference.EntityReferenceProvider;
import org.xwiki.model.reference.EntityReferenceResolver;
import org.xwiki.model.reference.EntityReferenceSerializer;
import org.xwiki.model.reference.LocalDocumentReference;
import org.xwiki.model.reference.SpaceReference;
import org.xwiki.query.Query;
import org.xwiki.query.QueryManager;
import org.xwiki.refactoring.RefactoringException;
import org.xwiki.refactoring.internal.job.PermanentlyDeleteJob;

import com.xpn.xwiki.XWiki;
import com.xpn.xwiki.XWikiContext;
import com.xpn.xwiki.XWikiException;
import com.xpn.xwiki.api.DeletedDocument;
import com.xpn.xwiki.doc.XWikiDeletedDocument;
import com.xpn.xwiki.doc.XWikiDocument;
import com.xpn.xwiki.internal.parentchild.ParentChildConfiguration;
import com.xpn.xwiki.store.XWikiRecycleBinStoreInterface;

/**
 * Default implementation of {@link ModelBridge} based on the old XWiki model.
 *
 * @version $Id: 51160e8212df4b0b2076b2c43e19ad8ea7a893a5 $
 * @since 7.4M2
 */
@Component
@Singleton
public class DefaultModelBridge implements ModelBridge
{
    /**
     * Regular expression used to match the special characters supported by the like HQL operator (plus the escaping
     * character).
     */
    private static final Pattern LIKE_SPECIAL_CHARS = Pattern.compile("([%_/])");

    /**
     * The reference to the type of object used to create an automatic redirect when renaming or moving a document.
     */
    private static final LocalDocumentReference REDIRECT_CLASS_REFERENCE =
        new LocalDocumentReference(XWiki.SYSTEM_SPACE, "RedirectClass");

    @Inject
    private Logger logger;

    /**
     * Used to perform the low level operations on entities.
     */
    @Inject
    private Provider<XWikiContext> xcontextProvider;

    /**
     * Used to query the child documents.
     */
    @Inject
    private QueryManager queryManager;

    /**
     * Used to serialize a space reference in order to query the child documents.
     */
    @Inject
    @Named("local")
    private EntityReferenceSerializer<String> localEntityReferenceSerializer;

    /**
     * Used to serialize the redirect location.
     *
     * @see #createRedirect(DocumentReference, DocumentReference)
     */
    @Inject
    private EntityReferenceSerializer<String> defaultEntityReferenceSerializer;

    /**
     * Used to resolve the references of child documents.
     */
    @Inject
    @Named("explicit")
    private DocumentReferenceResolver<String> explicitDocumentReferenceResolver;

    /**
     * Used to create the minimum need parent reference.
     */
    @Inject
    @Named("compact")
    private EntityReferenceSerializer<String> compactEntityReferenceSerializer;

    /**
     * Use to get back a relative reference based on a compact string reference.
     */
    @Inject
    @Named("relative")
    private EntityReferenceResolver<String> relativeStringEntityReferenceResolver;

    @Inject
    private DocumentReferenceResolver<EntityReference> documentReferenceResolver;

    @Inject
    private JobProgressManager progressManager;

    @Inject
    private ParentChildConfiguration parentChildConfiguration;

    @Inject
    private EntityReferenceProvider entityReferenceProvider;

    @Inject
    private Provider<LinkStore> linkStoreProvider;

    @Override
    public boolean create(DocumentReference documentReference)
    {
        XWikiContext xcontext = this.xcontextProvider.get();

        try {
            XWikiDocument newDocument = xcontext.getWiki().getDocument(documentReference, xcontext);
            xcontext.getWiki().saveDocument(newDocument, xcontext);
            this.logger.info("Document [{}] has been created.", documentReference);
            return true;
        } catch (Exception e) {
            this.logger.error("Failed to create document [{}].", documentReference, e);
            return false;
        }
    }

    @Override
    public boolean copy(DocumentReference source, DocumentReference destination)
    {
        XWikiContext xcontext = this.xcontextProvider.get();
        try {
            String language = source.getLocale() != null ? source.getLocale().toString() : null;
            boolean result =
                xcontext.getWiki().copyDocument(source, destination, language, false, true, true, xcontext);
            if (result) {
                this.logger.info("Document [{}] has been copied to [{}].", source, destination);
            } else {
                this.logger.warn(
                    "Cannot fully copy [{}] to [{}] because an orphan translation" + " exists at the destination.",
                    source, destination);
            }
            return result;
        } catch (Exception e) {
            this.logger.error("Failed to copy [{}] to [{}].", source, destination, e);
            return false;
        }
    }

    @Override
    public boolean delete(DocumentReference reference)
    {
        return delete(reference, false);
    }

    @Override
    public boolean delete(DocumentReference reference, boolean skipRecycleBin)
    {
        XWikiContext xcontext = this.xcontextProvider.get();
        try {
            XWikiDocument document = xcontext.getWiki().getDocument(reference, xcontext);
            if (document.getTranslation() == 1) {
                xcontext.getWiki().deleteDocument(document, !skipRecycleBin, xcontext);
                this.logger
                    .info("Document [{}] has been deleted (to the recycle bin: [{}]).", reference, !skipRecycleBin);
            } else {
                xcontext.getWiki().deleteAllDocuments(document, !skipRecycleBin, xcontext);
                this.logger
                    .info("Document [{}] has been deleted with all its translations (to the recycle bin: [{}]).",
                        reference,
                        !skipRecycleBin);
            }
            return true;
        } catch (Exception e) {
            this.logger
                .error("Failed to delete document [{}] (to the recycle bin: [{}]).", reference, !skipRecycleBin, e);
            return false;
        }
    }

    @Override
    public boolean removeLock(DocumentReference reference)
    {
        XWikiContext xcontext = this.xcontextProvider.get();
        try {
            XWikiDocument document = xcontext.getWiki().getDocument(reference, xcontext);

            if (document.getLock(xcontext) != null) {
                document.removeLock(xcontext);
                this.logger.info("Document [{}] has been unlocked.", reference);
            }

            return true;
        } catch (Exception e) {
            // Just warn, since it's a recoverable situation.
            this.logger.warn("Failed to unlock document [{}].", reference, e);
            return false;
        }
    }

    @Override
    public void createRedirect(DocumentReference oldReference, DocumentReference newReference)
    {
        XWikiContext xcontext = this.xcontextProvider.get();
        DocumentReference redirectClassReference =
            new DocumentReference(REDIRECT_CLASS_REFERENCE, oldReference.getWikiReference());
        try {
            if (xcontext.getWiki().exists(redirectClassReference, xcontext)) {
                XWikiDocument oldDocument = xcontext.getWiki().getDocument(oldReference, xcontext);
                int number = oldDocument.createXObject(redirectClassReference, xcontext);
                String location = this.defaultEntityReferenceSerializer.serialize(newReference);
                oldDocument.getXObject(redirectClassReference, number).setStringValue("location", location);
                oldDocument.setHidden(true);
                xcontext.getWiki().saveDocument(oldDocument, "Create automatic redirect.", xcontext);
                this.logger.info("Created automatic redirect from [{}] to [{}].", oldReference, newReference);
            } else {
                this.logger.warn("We can't create an automatic redirect from [{}] to [{}] because [{}] is missing.",
                    oldReference, newReference, redirectClassReference);
            }
        } catch (XWikiException e) {
            this.logger.error("Failed to create automatic redirect from [{}] to [{}].", oldReference, newReference, e);
        }
    }

    @Override
    public boolean canOverwriteSilently(DocumentReference documentReference)
    {
        try {
            XWikiContext xcontext = this.xcontextProvider.get();
            XWikiDocument document = xcontext.getWiki().getDocument(documentReference, xcontext);
            DocumentReference redirectClassReference =
                new DocumentReference(REDIRECT_CLASS_REFERENCE, documentReference.getWikiReference());
            // Overwrite silently the redirect pages.
            return document.getXObject(redirectClassReference) != null;
        } catch (XWikiException e) {
            this.logger.warn("Failed to get document [{}]. Root cause: [{}].", documentReference,
                ExceptionUtils.getRootCauseMessage(e));
            return false;
        }
    }

    @Override
    public boolean exists(DocumentReference reference) throws Exception
    {
        XWikiContext xcontext = this.xcontextProvider.get();
        return xcontext.getWiki().exists(reference, xcontext);
    }

    @Override
    public Set<DocumentReference> getBackLinkedDocuments(EntityReference reference) throws RefactoringException
    {
        Set<EntityReference> references;
        try {
            references = this.linkStoreProvider.get().resolveBackLinkedEntities(reference);
        } catch (Exception e) {
            throw new RefactoringException("Failed to resolve backlinks for entity [" + reference + "]", e);
        }

        XWikiContext xcontext = this.xcontextProvider.get();
        Set<DocumentReference> documentReferences = new HashSet<>(references.size());
        for (EntityReference entityReference : references) {
            documentReferences.add(this.documentReferenceResolver.resolve(entityReference, xcontext));
        }

        return documentReferences;
    }

    @Override
    public List<DocumentReference> getDocumentReferences(SpaceReference spaceReference)
    {
        try {
            // At the moment we don't have a way to retrieve only the direct children so we select all the descendants.
            // This means we select all the documents from the specified space and from all the nested spaces.
            String statement = "select distinct(doc.fullName) from XWikiDocument as doc "
                + "where doc.space = :space or doc.space like :spacePrefix escape '/'";
            Query query = this.queryManager.createQuery(statement, Query.HQL);
            query.setWiki(spaceReference.getWikiReference().getName());
            String localSpaceReference = this.localEntityReferenceSerializer.serialize(spaceReference);
            query.bindValue("space", localSpaceReference);
            String spacePrefix = LIKE_SPECIAL_CHARS.matcher(localSpaceReference).replaceAll("/$1");
            query.bindValue("spacePrefix", spacePrefix + ".%");

            List<DocumentReference> descendants = new ArrayList<>();
            for (Object fullName : query.execute()) {
                descendants.add(this.explicitDocumentReferenceResolver.resolve((String) fullName, spaceReference));
            }
            return descendants;
        } catch (Exception e) {
            this.logger.error("Failed to retrieve the documents from [{}].", spaceReference, e);
            return Collections.emptyList();
        }
    }

    @Override
    public boolean updateParentField(final DocumentReference oldParentReference,
        final DocumentReference newParentReference)
    {
        XWikiContext context = xcontextProvider.get();
        XWiki wiki = context.getWiki();

        boolean popLevelProgress = false;
        try {
            // Note: This operation could have been done in Hibernate (using the Store API) in one single update query.
            // However, due to XWiki's document cache, it`s better in the end to use the Document API and update each
            // child document individually.
            XWikiDocument oldParentDocument = wiki.getDocument(oldParentReference, context);

            List<DocumentReference> childReferences = oldParentDocument.getChildrenReferences(context);

            if (childReferences.size() > 0) {
                this.progressManager.pushLevelProgress(childReferences.size(), this);
                popLevelProgress = true;
            }

            for (DocumentReference childReference : childReferences) {
                this.progressManager.startStep(this);

                XWikiDocument childDocument = wiki.getDocument(childReference, context);
                childDocument.setParentReference(newParentReference);

                wiki.saveDocument(childDocument, "Updated parent field.", true, context);

                this.progressManager.endStep(this);
            }

            if (childReferences.size() > 0) {
                this.logger.info("Document parent fields updated from [{}] to [{}] for [{}] documents.",
                    oldParentReference, newParentReference, childReferences.size());
            }
        } catch (Exception e) {
            this.logger.error("Failed to update the document parent fields from [{}] to [{}].", oldParentReference,
                newParentReference, e);
            return false;
        } finally {
            if (popLevelProgress) {
                this.progressManager.popLevelProgress(this);
            }
        }

        return true;
    }

    @Override
    public DocumentReference setContextUserReference(DocumentReference userReference)
    {
        XWikiContext context = xcontextProvider.get();
        DocumentReference previousUserReference = context.getUserReference();
        context.setUserReference(userReference);
        return previousUserReference;
    }

    @Override
    public void update(DocumentReference documentReference, Map<String, String> parameters)
    {
        try {
            XWikiContext context = xcontextProvider.get();
            XWiki wiki = context.getWiki();
            XWikiDocument document = wiki.getDocument(documentReference, context);
            boolean save = false;

            String title = parameters.get("title");
            if (title != null && !title.equals(document.getTitle())) {
                document.setTitle(title);
                save = true;
            }

            // Some old applications still rely on the parent/child links between documents.
            // For the retro-compatibility, we synchronize the "parent" field of the document with the (real)
            // hierarchical parent.
            //
            // But if the user has voluntary enabled the legacy "parent/child" mechanism for the breadcrumbs, we keep
            // the old behaviour when location and parent/child mechanism were not linked.
            //
            // More information: https://jira.xwiki.org/browse/XWIKI-13493
            if (!parentChildConfiguration.isParentChildMechanismEnabled()) {

                // we compute the new hierarchical parent
                DocumentReference hierarchicalParent = getHierarchicalParent(documentReference);

                // we compute a relative reference for the hierarchical parent
                String hierarchicalParentSerialized =
                    this.compactEntityReferenceSerializer.serialize(hierarchicalParent, documentReference);
                EntityReference relativeHierarchicalReference = this.relativeStringEntityReferenceResolver
                    .resolve(hierarchicalParentSerialized, EntityType.DOCUMENT);

                // we can rely on the current document parent ref, as after the copy the encoded parent is not changed.
                EntityReference newDocumentParentRef = document.getRelativeParentReference();

                // all cases are supported:
                //   1. if we move on a same wiki but on a different space: the relative ref are different, the
                //      parent will be updated
                //   2. if we move on a same wiki, same space: the relative ref are the same, we keep the parent
                //   3. if we move on a different wiki, different space: we don't care about the wiki, we are in the
                //      same case as 1
                //   4. if we move on a different wiki, same space: actually we can keep the old parent relative
                //      reference, as it's relative of the document reference. So the parent reference will be well
                //      computed with the already existing relative reference.
                if (!relativeHierarchicalReference.equals(newDocumentParentRef)) {
                    document.setParentReference(relativeHierarchicalReference);
                    save = true;
                }
            }

            if (save) {
                wiki.saveDocument(document, "Update document after refactoring.", true, context);
                this.logger.info("Document [{}] has been updated.", documentReference);
            }
        } catch (Exception e) {
            this.logger.error("Failed to update the document [{}] after refactoring.", documentReference, e);
        }
    }

    private DocumentReference getHierarchicalParent(DocumentReference documentReference)
    {
        final String spaceHomePage = entityReferenceProvider.getDefaultReference(EntityType.DOCUMENT).getName();

        EntityReference parentOfTheSpace = documentReference.getLastSpaceReference().getParent();

        boolean pageIsNotTerminal = documentReference.getName().equals(spaceHomePage);

        // Case 1: The document has the location A.B.C.WebHome
        // The parent should be A.B.WebHome
        if (pageIsNotTerminal && parentOfTheSpace.getType() == EntityType.SPACE) {
            return new DocumentReference(spaceHomePage, new SpaceReference(parentOfTheSpace));
        }

        // Case 2: The document has the location A.WebHome
        // The parent should be Main.WebHome
        if (pageIsNotTerminal && parentOfTheSpace.getType() == EntityType.WIKI) {
            return new DocumentReference(spaceHomePage,
                new SpaceReference(entityReferenceProvider.getDefaultReference(EntityType.SPACE).getName(),
                    documentReference.getWikiReference()));
        }

        // Case 3: The document has the location A.B
        // The parent should be A.WebHome
        return new DocumentReference(spaceHomePage, documentReference.getLastSpaceReference());
    }

    private XWikiDeletedDocument getDeletedDocument(XWikiContext context, long deletedDocumentId) throws XWikiException
    {
        XWiki xWiki = context.getWiki();
        XWikiDeletedDocument deletedDocument = xWiki.getDeletedDocument(deletedDocumentId, context);
        if (deletedDocument == null) {
            logger.error("Deleted document with ID [{}] does not exist.", deletedDocumentId);
            return null;
        }
        return deletedDocument;
    }

    @Override
    public boolean restoreDeletedDocument(long deletedDocumentId, AbstractCheckRightsRequest request)
    {
        XWikiContext context = this.xcontextProvider.get();
        XWiki xwiki = context.getWiki();

        DocumentReference deletedDocumentReference = null;
        try {
            // Retrieve the deleted document by ID.
            XWikiDeletedDocument deletedDocument = this.getDeletedDocument(context, deletedDocumentId);
            if (deletedDocument == null) {
                return false;
            }
            deletedDocumentReference = deletedDocument.getDocumentReference();

            // If the document (or the translation) that we want to restore does not exist, restore it.
            if (xwiki.exists(deletedDocumentReference, context)) {
                // TODO: Add overwrite support maybe also with interactive (question/answer) mode.
                // Default for now is to skip and log as error to restore over existing documents.
                logger.error("Document [{}] with ID [{}] can not be restored. Document already exists",
                    deletedDocument.getFullName(), deletedDocumentId);
            } else if (request.isCheckAuthorRights()
                && !canRestoreDeletedDocument(deletedDocument, context.getAuthorReference())) {
                logger.error("The author [{}] of this script is not allowed to restore document [{}] with ID [{}]",
                    context.getAuthorReference(), deletedDocumentReference, deletedDocumentId);
            } else if (request.isCheckRights()
                && !canRestoreDeletedDocument(deletedDocument, context.getUserReference())) {
                logger.error("You are not allowed to restore document [{}] with ID [{}]", deletedDocumentReference,
                    deletedDocumentId);
            } else {
                // Restore the document.
                xwiki.restoreFromRecycleBin(deletedDocument.getId(), "Restored from recycle bin", context);

                logger.info("Document [{}] has been restored", deletedDocumentReference);

                return true;
            }
        } catch (Exception e) {
            // Try to log the document reference since it`s more useful than the ID.
            if (deletedDocumentReference != null) {
                logger.error("Failed to restore document [{}] with ID [{}]", deletedDocumentReference,
                    deletedDocumentId, e);
            } else {
                logger.error("Failed to restore deleted document with ID [{}]", deletedDocumentId, e);
            }
        }

        return false;
    }

    @Override
    public List<Long> getDeletedDocumentIds(String batchId)
    {
        XWikiContext context = this.xcontextProvider.get();
        XWiki xwiki = context.getWiki();

        List<Long> result = new ArrayList<>();
        try {
            XWikiDeletedDocument[] deletedDocuments =
                xwiki.getRecycleBinStore().getAllDeletedDocuments(batchId, false, context, true);
            for (XWikiDeletedDocument deletedDocument : deletedDocuments) {
                result.add(deletedDocument.getId());
            }
        } catch (Exception e) {
            logger.error("Failed to get deleted document IDs for batch [{}]", batchId);
        }

        return result;
    }

    protected boolean canRestoreDeletedDocument(XWikiDeletedDocument deletedDocument, DocumentReference userReference)
    {
        boolean result = false;

        XWikiContext context = this.xcontextProvider.get();

        // Remember the context user.
        DocumentReference currentUserReference = context.getUserReference();
        try {
            // Reuse the DeletedDocument API to check rights.
            DeletedDocument deletedDocumentApi = new DeletedDocument(deletedDocument, context);

            // Note: DeletedDocument API works with the current context user.
            context.setUserReference(userReference);

            result = deletedDocumentApi.canUndelete();
        } catch (Exception e) {
            logger.error("Failed to check restore rights on deleted document [{}] for user [{}]",
                deletedDocument.getId(), userReference, e);
        } finally {
            // Restore the context user;
            context.setUserReference(currentUserReference);
        }

        return result;
    }

    protected boolean canPermanentlyDeleteDocument(XWikiDeletedDocument deletedDocument,
        DocumentReference userReference)
    {
        boolean result = false;

        XWikiContext context = this.xcontextProvider.get();

        // Remember the context user.
        DocumentReference currentUserReference = context.getUserReference();
        try {
            // Reuse the DeletedDocument API to check rights.
            DeletedDocument deletedDocumentApi = new DeletedDocument(deletedDocument, context);

            // Note: DeletedDocument API works with the current context user.
            context.setUserReference(userReference);

            result = deletedDocumentApi.canDelete();
        } catch (Exception e) {
            logger.error("Failed to check delete rights on deleted document [{}] for user [{}]",
                deletedDocument.getId(), userReference, e);
        } finally {
            // Restore the context user;
            context.setUserReference(currentUserReference);
        }

        return result;
    }

    @Override
    public boolean permanentlyDeleteDocument(long deletedDocumentId, AbstractCheckRightsRequest request)
    {
        XWikiContext context = this.xcontextProvider.get();
        XWiki xwiki = context.getWiki();

        DocumentReference deletedDocumentReference = null;
        try {
            // Retrieve the deleted document by ID.
            XWikiDeletedDocument deletedDocument = this.getDeletedDocument(context, deletedDocumentId);

            if (deletedDocument == null) {
                return false;
            }
            deletedDocumentReference = deletedDocument.getDocumentReference();

            if (request.isCheckRights() && !canPermanentlyDeleteDocument(deletedDocument, context.getUserReference())) {
                logger.error("You are not allowed to permanently delete document [{}] with ID [{}]",
                    deletedDocumentReference, deletedDocumentId);
            } else if (request.isCheckAuthorRights()
                && !canPermanentlyDeleteDocument(deletedDocument, context.getAuthorReference())) {
                logger.error("The author [{}] of this script is not allowed to permanently deleted document [{}] with "
                        + "id [{}]", context.getAuthorReference(), deletedDocumentReference, deletedDocumentId);
            } else {
                // Restore the document.
                xwiki.getRecycleBinStore().deleteFromRecycleBin(deletedDocumentId, context, true);

                logger.info("Document [{}] has been permanently deleted.", deletedDocumentReference);

                return true;
            }
        } catch (Exception e) {
            // Try to log the document reference since it`s more useful than the ID.
            if (deletedDocumentReference != null) {
                logger.error("Failed to permanently delete document [{}] with ID [{}]", deletedDocumentReference,
                    deletedDocumentId, e);
            } else {
                logger.error("Failed to permanently delete document with ID [{}]", deletedDocumentId, e);
            }
        }

        return false;
    }

    @Override
    public boolean permanentlyDeleteAllDocuments(PermanentlyDeleteJob deleteJob, AbstractCheckRightsRequest request)
    {
        XWikiContext context = this.xcontextProvider.get();
        XWiki xwiki = context.getWiki();
        int limit = 10;

        XWikiRecycleBinStoreInterface recycleBinStore = xwiki.getRecycleBinStore();
        try {
            long numberOfDocumentsToDelete = recycleBinStore.getNumberOfDeletedDocuments(context);

            int nbDocToDelete;
            if (numberOfDocumentsToDelete > Integer.MAX_VALUE) {
                logger.warn("Only [{}] file can be deleted at once. Please run again the job to delete everything.",
                    Integer.MAX_VALUE);
                nbDocToDelete = Integer.MAX_VALUE;
            } else {
                nbDocToDelete = Integer.valueOf(numberOfDocumentsToDelete + "");
            }
            progressManager.pushLevelProgress(nbDocToDelete, deleteJob);

            for (int i = 0; i < nbDocToDelete; i += limit) {
                Long[] allDeletedDocumentsIds = recycleBinStore.getAllDeletedDocumentsIds(context, limit);
                for (Long deletedDocumentsId : allDeletedDocumentsIds) {
                    if (deleteJob.getStatus().isCanceled()) {
                        return false;
                    } else {
                        this.progressManager.startStep(deleteJob);
                        this.permanentlyDeleteDocument(deletedDocumentsId, request);
                        this.progressManager.endStep(deleteJob);
                    }
                }
            }
            return true;
        } catch (XWikiException e) {
            logger.error("Failed to permanently delete all documents", e);
        }

        return false;
    }

    @Override
    public boolean rename(DocumentReference source, DocumentReference destination)
    {
        XWikiContext xcontext = this.xcontextProvider.get();
        try {
            return xcontext.getWiki()
                .renameDocument(source, destination, true,
                    Collections.emptyList(), Collections.emptyList(), xcontext);
        } catch (Exception e) {
            this.logger.error("Failed to rename [{}] to [{}].", source, destination, e);
        }
        return false;
    }
}
